package sam.machine.engine ; 

import java.util.* ;
import java.io.* ; 
import sam.machine.lang.* ;
import sam.util.* ;
import sam.gui.search.* ;
import java.awt.* ;
import javax.swing.* ;  
import java.util.concurrent.*; 

/**    
  *  TangleTool  
  *  Tests designed to study how to retrieve branch terms
  *  for rule satisfaction and equality modulation 
  *  for an autolog Skolem Machine program.   
  *  ATP (algebraic term profile) storage and retrieval design testkit.
  *  Goal is to design/test active components for sam.machine.engine.
  *  @version 4/22/2022  ModTest (locate/match)
  *  @author John R. Fisher 
   */ 
public class TangleTool {
   /**  
     * Read/compile a branch and a tangle test query from AutoLogEdit
     *   a theory-branch rule, e.g.
     *      true =>               // branch[0]
     *         p(a,b),            // branch[1]   
     *         p(a,f(a)),         // branch[2]    
     *         a=c.               // branch[3]  
     *   a tangle query, e.g. 
     *    % p(X,b) => fact.      // match literal p(X,b) with  branch facts
     *    % f(a)=b.              // modulate branch facts using rewrite f(a)=b
     *    % f(X)=g(X).           // modulate branch facts using rewrite f(X)=g(X)
      */
   protected Theory theory ;   
   // computed by constructor using maxArity(theory), for ALL rules and rewrites
   protected int maxArity ;     
   protected Functor[] BRANCH ;     // constructed from first inference Rule form
   protected Inference query ;   // the second input form (Rule or Rewrite)
   protected static int OFFSET = 3 ; // colog14I used 5
   // key profile source, MAX of 10 args for TangleTool testing
   protected boolean[][] KEY = new boolean[1024][10] ; 

   /* 
      FactMap: ConcurrentHashmap retrieval via algebraic term profiles for branch facts.
      To be ducumented in Autolog Design Notebook Chapter 6.
   */
   protected ConcurrentHashMap<Integer,Vector<Integer>> FactMap ;  
   
   // power=2^arity         arity: 0 1 2 3  4  5  6  7   8   9   10
   protected static int[] power = {1,2,4,8,16,32,64,128,256,512,1024} ;  
   // accelerator for pre-computation of KEY, called by TangleTool constructor
   // See END NOTE #2.  TangleTool over-fills FactMap in order to test 
   // for map clashes

   protected void setKEY() {   
      // row 0 is all false
      for (int row=1 ; row < 1024 ; row++)  {
         // increment the key pattern
         boolean carry = true ;  // increment last row
         for (int col=0; col < 10; col++) {
            if (!KEY[row-1][col]) {  // previous row key
               if (carry) {
                  KEY[row][col]=true ;
                  carry=false ;
               }  
               else KEY[row][col]=false ;
            }
            else { // KEY[row-1][col]==true
               if (carry) {  
                  KEY[row][col]=false ;
                  carry=true ;
               } 
               else KEY[row][col]=true ;
            }
         }
      }
   }
  
   /**
     * Check to see if a theory generates an acceptable test 
     * before constructing TangleTool.  
      */
   public static boolean check(Theory theory) { 
      if (theory.isEmpty() || theory.size() != 2  
                           || !theory.isIndexical()) // remove?? 
         return false ;
      // branch rule
      Rule rule = (Rule)(theory.elementAt(0)) ;
      if (rule.antecedent.size()>1)  return false ; 
      if (rule.consequent.size()>1)  return false ; 
      Term t = rule.antecedent.elementAt(0) ;
      if (!t.name.equals("true") | t.arity!=0 )  return false ;    
      Conjunction c = (Conjunction)(rule.consequent.elementAt(0)) ;   
      if (!c.factual)  return false ;   // via AutoLogAnalyzer for theory  
      // query 
      Inference query = (Inference)(theory.elementAt(1)) ;   
      if (query instanceof Rule) {  // check form of rule
         Rule qrule = (Rule)query ;
         if (qrule.consequent.size() > 1) return false ;
         Conjunction qc = (Conjunction)(qrule.consequent.elementAt(0)) ; 
         if (qc.size() > 1) return false ;
         Term qt = qc.elementAt(0) ;
         if (!qt.name.equals("fact") | qt.arity!=0) return false ;
         if (qrule.antecedent.size() > 1) return false ;
      } 
      // otherwise query is a Rewrite component and already ✔ checked
      return true ;
   }

   /** 
     *  Construct tool components assuming ✔ checked, 
     *  TangleTool may need to know auxArea of Viewer
      */
   public TangleTool(Theory thy) { 
      // sam.gui.search.Autolog.java has protected theory
      // which is "shared" with this TangleTool after checked:
      this.theory = thy ;  
      this.maxArity = maxArity(this.theory) ;  
      Rule rule = (Rule)this.theory.elementAt(0) ;
      Conjunction ant = (Conjunction)(rule.antecedent) ;
      Conjunction cons = rule.consequent.elementAt(0) ; 
      this.BRANCH = new Functor[1+cons.size()] ; 
      BRANCH[0] = ant.elementAt(0) ;  // true root
      // rest of branch fact terms ...
      for (int i=0; i<cons.size(); i++) 
         this.BRANCH[i+1] = cons.elementAt(i) ; 
      // extract the query
      this.query = (Inference)this.theory.elementAt(1) ; 
      // accelerate the KEY component
      this.setKEY() ;
      this.FactMap = new ConcurrentHashMap<Integer,Vector<Integer>>() ; 
//      this.ModMap = new ConcurrentHashMap<Integer,Vector<Vector<Integer>>>() ; 
      // The default load factor of a HashMap is 0.75f.
   }

   // largest functor arity in a theory
   // Suggest CANCEL tests if maxArity > 10
   // Production runtime could allow larger arity
   // using various to-be-decided adaptations.
   protected int maxArity(Theory thy) { 
      int n=0 ; 
      for(int j=0 ; j < theory.arity.length ; j++) 
         if (thy.arity[j] > n) n = thy.arity[j] ;
      return n ;
   } 

   /**                   
   *   atpHashCode calculations for a GROUNDED branch fact
   *   using a particular bound key pattern. 
   *   N.B., args terms are Functor terms.
   *   Use KEY as key accelerator in remember.
    */
   public static int atpHashCode(Functor f, boolean[] key) {
      int h ; 
      if (f.fnctr != null)   // f parametric
         h = atpHashCode(f.fnctr) ;
      else h = f.index ;     // f named
      for (int i = 0 ; i < f.arity ; i++) // use key pattern
         h = (h<< OFFSET) + (key[i] ? atpHashCode(f.args[i]) : 0) ; 
      return h ;        //   ^ N.B. NOT f.key[i] 
   }
   /** For embedded ground terms Embedded terms */
   public static int atpHashCode(Term t) { 
      Functor f = (Functor)t ;  
      int hc ; 
      if (f.fnctr != null)   // f parametric
         hc = atpHashCode(f.fnctr) ;
      else hc = f.index ;     // f named
      for (int i = 0 ; i < f.arity ; i++) { // can accelerate?
         hc = (hc << OFFSET) + atpHashCode(f.args[i]) ;
      } 
      return hc ; 
   }

   /** 
     *  Remember a bound facts's branch location (bloc) 
     *  in FactMap. This (over) remembers ALL key profiles,
     *  and is a test for map clashes. Possibility for clashes   
     *  is reason for code " if (!locations.contains(bloc)) ",
     *  Production remembering may use analysed key profiles 
     *  where appropriate.
      */
   protected void rememberFact(Functor f, int bloc) { 
      int arity = f.arity ; 
      if (arity == 0) { // a named predicate or constant arg
         Vector<Integer> locations = FactMap.get(f.index) ; 
         if (locations == null) {
            Vector<Integer> vec = new Vector<Integer>() ;
            vec.add(bloc) ;
            FactMap.put(f.index, vec) ; 
         }
         else if (!locations.contains(bloc))
               locations.add(bloc) ; 
         else {} // bloc already recorded
         return ; 
      }
      else { // arity ≥ 1  so assert to FactMap by key 
         for (int row=0 ; row < this.power[arity] ; row++) {    
            boolean[] key = this.KEY[row] ;
            int hcode = atpHashCode(f,key) ; 
            Vector<Integer> locations = FactMap.get(hcode) ;
            if (locations == null) { 
               Vector<Integer> vec = new Vector<Integer>() ;
               vec.add(bloc) ;
               FactMap.put(hcode, vec) ; 
            }
            else if (!locations.contains(bloc)) { 
               locations.add(bloc) ;  
            }
            else {} // bloc already recorded
         }
      }
   }   

   /**
     * Describe internal subterm locations for a Functor f.
     * Discriminate all possible subterms by precise location. 
     * Start with path=[b,...] for BRANCH[b] and initialized empty buffer.
     * Buffer generates the location discrimination report.
      */
   public static void describe(Functor f, Vector<Integer> path, StringBuffer buf) { 
      buf.append (f) ; 
      buf.append("  ") ;
      buf.append("@"+path) ;
      buf.append("\n") ; 
      // describe f.fnctr if not null
      if(f.fnctr != null) { 
         // start a new path
         //Vector<Integer> newpath = Vector<Integer>path.clone() ;
         Vector<Integer> newpath = copy(path) ;  // no warnings?
         newpath.add(-1) ;
         describe((Functor)f.fnctr, newpath, buf) ; 
      }
      // describe all f.args
      for (int i=0 ; i < f.arity ; i++) {
         // start a new path
         //Vector<Integer> newpath = Vector<Integer>path.clone() ;
         Vector<Integer> newpath = copy(path) ;
         newpath.add(i) ; 
         describe((Functor)f.args[i], newpath,buf) ;
      }
   }
   /*
   Method describe is a demonstration experiment to express a
   descrimination network representation for all possible subterms for 
   a branch fact. The query needs to be acceptable modulation test. 
   Use describe to help guide design for locate.
      eg. f=p(a)(f(a),g(a)) @branch fact # b
      desc(f):  
         p(a)(f(a),g(a)) @[b]
         p(a)            @[b,-1]
         a               @[b,-1,0]
         f(a)            @[b,0]
         a               @[b,0,0] 
         g(a)            @[b,1]
         a               @[b,1,0] 
      3/18/2022 finally got this to work as intended.
   */ 

   /**   
    * Quick copy of previous term locator to use 
    * for a next locator. .clone() works but warnings persist.
     */
   public static Vector<Integer> copy(Vector<Integer> in) {
      Vector<Integer> out = new Vector<Integer>(in.size()+1) ;
      Iterator it = in.iterator() ;
      while (it.hasNext()) out.add((Integer)it.next()) ;
      return out ;
   } 

   /**  
     *  Locate (acceptible) lterm matches in a branch fact f
     +  using design approach with match                                             
      */
   public static void locate(Functor lterm,                // lterm= -.
                             int numvars,                  // #vars in lterm
                             Functor f,                    // BRANCH[i] 
                             Vector<Integer> path,         // [i....] 
                             Vector<Vector<Integer>> locs, // [[..], ...]
                             Vector<Functor[]> bindings       // [Functor[], ...]
                                                      ){ 
      // are lterm and f matchable? 
      Functor[] varbind = new Functor[numvars] ;  // for this attempt
      if (match(lterm,f,varbind)) {
         locs.add(path) ; 
         bindings.add(varbind) ;
      }
      // locate lterm in f.fnctr
      if(f.fnctr != null) { 
         // start a new path
         //Vector<Integer> newpath = Vector<Integer>path.clone() ;
         Vector<Integer> newpath = copy(path) ;  // no warnings?
         newpath.add(-1) ;
         locate(lterm, numvars, (Functor)f.fnctr, newpath, locs, bindings) ; 
      }
      // locate lterm in f.args
      for (int i=0 ; i < f.arity ; i++) {
         // start a new path
         //Vector<Integer> newpath = Vector<Integer>path.clone() ;
         Vector<Integer> newpath = copy(path) ;
         newpath.add(i) ;    
         locate(lterm, numvars, (Functor)f.args[i], newpath,locs, bindings) ;
      }
   }

   /**  
     *  Does left match bound Functor f using bindings from varbind? 
     *  If so then match returns true and f can be modulated
     *  using left=right via varbind at active inference time. 
     *  A call for match should use an initialized 3rd parameter. 
     *   For example,
     *    lterm     f
     *   h(X,Y)  h(a,f(b)) -> true, varbind[0]=a,  varbind[1]=f(b)
      */
   public static boolean match(Term left, Functor f, Functor[] varbind) { 
      //     Variable or Functor ^          ^ bound Functor
      if (left instanceof Variable) { 
         // use lookup value 
         if (varbind[left.index] == null) {
            varbind[left.index] = f ; 
            return true ; 
         } 
         else return match(varbind[left.index],f,varbind) ; 
      } // end left is Variable
      // otherwise left is Functor
      Functor leFt = (Functor)left ;  // CAST left to Functor
      if (leFt.arity != f.arity) return false ;  
      if (f.fnctr != null) { // f is parametric 
         if (leFt.fnctr == null) return false ;
         // otherwise both are parametric
         if (!match(leFt.fnctr,(Functor)f.fnctr,varbind)) return false ;  
         // leFt.fnctr & f.fnctr match
         // now match args 
         for(int i=0; i<left.arity; i++)
            if (!match(leFt.args[i],(Functor)f.args[i],varbind)) return false ;
         return true ;
      } // end both are parametric 
      // otherwise both are named    // debug w/tuff0
      if (f.index != leFt.index) return false ; // different names
      // names match, so now match args
      for (int i=0; i<leFt.arity; i++)  
         if (!match(leFt.args[i],(Functor)f.args[i],varbind)) return false ; 
      return true ; 
   }
   /* N.B. consider to allow predicative parametric modulator like  p(X)(Y)=q(X,Y).
        OK, using current locate ideas
       literals?  true => true => p(a)(b)(c), q(a). 
                  p(a)(X)(Y) => fact.  ??? Probably not. 
   */


//<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><><>
   /** 
     *  showTest()
     *  branch and query setup  
     *  Working with maxArity, 2/27/2022
      */
   public String showTest() {   // branch and query setup
      StringBuffer sb = new StringBuffer() ; 
      sb.append("TangleTool Test Problem \n") ; 
      sb.append("BRANCH \n") ; 
      for (int i=0 ; i < this.BRANCH.length ; i++) {
         sb.append("   "+ i + " ") ;
         sb.append(this.BRANCH[i] +"\n") ;
      }
      sb.append("TANGLE TEST:  \n") ;  
      if (query instanceof Rule) { 
         Term literal = ((Rule)query).antecedent.elementAt(0) ;
         sb.append("Find matches for literal "+literal+" with branch facts.\n") ;
      } 
      else { // query instance of Rewrite
         Rewrite rewrite = (Rewrite)query ; 
         sb.append("{Application of rewrite "+ rewrite.Lterm +"="+rewrite.Rterm+" to BRANCH[1].}\n") ;
      } 
      // WARN if maxArity>10 -- should CANCEL test. 
      // TangleTool restricts maxArity≤10, because all possible keys are mapped (to test for clashes).
      // Production active runtime components may relax this testing restriction.
      sb.append("maxArity="+this.maxArity+ (this.maxArity>10 ? " TOO LARGE(CANCEL) \n"  : "\n")) ;
      return sb.toString() ; 
   }

   /**
     *  LAUNCH TangleTool QUERY for FactMapTest or ModTest
     *  TangleTool query widget 
     *  Viewer calls construction of TangleTool before widget displayed.
      */
   public void showQueryWidget(AutoLog viewer) { 
      // Set up for JOptionPane.showOptionDialog 
      Object[] options = { "OK", "CANCEL" };
      int result = 
         JOptionPane.showOptionDialog(null, "Click OK to continue", "TANGLE TEST",
             JOptionPane.DEFAULT_OPTION, JOptionPane.WARNING_MESSAGE,
             null, options, options[0]);
      if (result == 0) {  // code tests, see END NOTES re early tests ...  
      if (this.query instanceof Rule) 
          FactMapTest(viewer) ;  
      else ModTest(viewer) ; 
      }
   } 

   ////////////////////
   // FactMapTest  inspect FactMap.toString() after remembering branch facts, 
   //    and locate which branch[i]'s relevant for query 'F=>fact.'
   //    using F.key.
   // 
   ////////////////////
   protected void FactMapTest(AutoLog viewer) {   
      String test_statement = this.showTest() ;  
      for (int i=0 ; i < this.BRANCH.length ; i++) 
         rememberFact((Functor)this.BRANCH[i], i) ;  
      // isolate Functor literal, subject of this test
      Functor f = ((Rule)this.query).antecedent.elementAt(0) ;  
      int hashf = atpHashCode(f, f.key) ; // NOTE f.key
      int checksum = 0 ;
      for (int i=1 ; i<this.BRANCH.length ; i++)
             checksum+= Math.pow(2,this.BRANCH[i].arity) ;
      String check = 
             (FactMap.keySet().size()-1 == checksum ? " ✔ (no clashes)" : " ↯ (clashes)") ;
      StringBuffer buf = new StringBuffer() ;  
      buf.append(test_statement+"\n\n") ;
      buf.append("FactMap mappings ⇲ \n") ;
      buf.append(FactMap.toString()) ; // ;-) so simple 
      buf.append("\n\nFactMap.keyset().size()="+FactMap.keySet().size() + check ) ;
      buf.append("\n\n"+ f +"↦"+hashf+"="+ FactMap.get(hashf)) ;  
      viewer.auxArea.setText(String.valueOf(buf.toString())) ; 
   }   
   
   ////////////////////
   //   ModTest    Under Construction 
   //  Simulate actions for a single ActiveModulator using
   //  subterm locations and variable bindings.
   //  Note AutoLogAnalyzer version 2/9/2022:
   //  Rule keyPatterns and Lterm static rewriteKeyPatterns separated.
   ////////////////////
   protected void ModTest(AutoLog viewer) {   
      String test_statement = this.showTest() ;  
      Rewrite r = (Rewrite)this.query ;
      Functor left = (Functor)r.Lterm ; // cast
      Term    right = r.Rterm ;
      StringBuffer buf = new StringBuffer() ;  
      buf.append(test_statement+"\n\n") ;
      boolean[] key = left.key ;  
      buf.append("Lterm key profile="+string(key)+"\n") ;  
      // ----- describe -----
      buf.append("Describe subterm locations for [1]"+BRANCH[1]+"\n") ; 
      Functor f = BRANCH[1] ; 
      Vector<Integer> path = new Vector<Integer>() ;
      path.add(1) ;
      StringBuffer bfr = new StringBuffer() ;
      describe(f, path, bfr) ;    
      bfr.append("describe finished\n\n") ;
      buf.append(bfr) ;
      ////////////////////////////////////////////////////////////////
      // ----- match & locate -----
      Functor[] varbind = new Functor[r.U] ; 
      //// just match -- seems OK
           //buf.append("match(left,f,varbind)="+match(left,f,varbind)+"\n") ; 
           //buf.append(string(varbind)) ; 
      ///// locate ... ???
      /*     public static void locate(Functor lterm,                // lterm= -.
                             int numvars,                  // #vars in lterm
                             Functor f,                    // BRANCH[i] 
                             Vector<Integer> path,         // [i....] 
                             Vector<Vector<Integer>> locs, // [[..], ...]
                             Vector<Functor[]> bindings       // [Functor[], ...]
      */
      // initial inputs 
         // lterm <- left 
         // numvars <- r.U
         // f <- f
         // path <- path
      Vector<Vector<Integer>> locs = new Vector<Vector<Integer>>() ;
      Vector<Functor[]> bindings = new Vector<Functor[]>() ;
      locate(left,r.U,f,path,locs,bindings) ;  // void return
      // How display stuff?
      buf.append("# locations="+locs.size()+"\n") ;
      buf.append("Locations and bindings: \n") ;
      for (int i=0 ; i<locs.size() ; i++) {
         buf.append(locs.elementAt(i)) ;
         buf.append("\n") ;
         buf.append(string(bindings.elementAt(i))) ;
         buf.append("----------\n");
      }
      viewer.auxArea.setText(String.valueOf(buf.toString())) ; 

   }   

   // Display a varbind table 
   private static String string(Functor[] vb) { 
      StringBuffer sb = new StringBuffer() ;
      for (int i=0 ; i<vb.length ; i++) 
         sb.append(i + " ⇾ " + vb[i] + "\n") ;
      return sb.toString() ;
   }

   // Show boolean[] profile keys 
   //    e.g., [false,true] for Lterm h(a,Y)
   private static String string(boolean[] k) { 
      StringBuffer sb = new StringBuffer() ; 
      sb.append("[") ; 
      for(int i = 0 ; i < k.length ; i++) {
         sb.append(k[i]) ; 
         if (i < k.length -1) sb.append(",") ; 
     }
      sb.append("]") ; 
      return sb.toString() ; 
   }

   /** 
     *  Experiment to reveal how << shift works
     *  Java parses (k<<l + m) as k<<(l+m)  a.k.o right assoc
     *  and k<<(l+m) has more shift spread than (k<<l) + m
      */ 
   public static void main (String[] args) { 
      // true => p(a)(c,d).    in AutoLogEdit
      //         ^4^5 ^6^7 
      System.out.println  ( (((4<<5) + 5)<<5 + 6)<<5 + 7 ) ;    // 1115684864
      System.out.println  ( (((((4<<5) + 5)<<5) + 6)<<5)+ 7 ) ; // 136391
      //    (k<<l)        1 23456    6    5   4    3   2    1
   } 

//<><><><><><><><><><><><><><><><><><><><><><><><><><><><><><> 

} // end TangleTool  


